@@ -13,6 +13,45 @@ use std::{
 #[cfg(unix)]
 use std::os::unix::fs::PermissionsExt;
 
+/// File mode indicating execute bit for other.
+pub const S_IXOTH: u32 = 0o1;
+/// File mode indicating write bit for other.
+pub const S_IWOTH: u32 = 0o2;
+/// File mode indicating read bit for other.
+pub const S_IROTH: u32 = 0o4;
+/// File mode indicating execute bit for group.
+pub const S_IXGRP: u32 = 0o10;
+/// File mode indicating write bit for group.
+pub const S_IWGRP: u32 = 0o20;
+/// File mode indicating read bit for group.
+pub const S_IRGRP: u32 = 0o40;
+/// File mode indicating execute bit for owner.
+pub const S_IXUSR: u32 = 0o100;
+/// File mode indicating write bit for owner.
+pub const S_IWUSR: u32 = 0o200;
+/// File mode indicating read bit for owner.
+pub const S_IRUSR: u32 = 0o400;
+/// Sticky bit.
+pub const S_ISVTX: u32 = 0o1000;
+/// Set GID bit.
+pub const S_ISGID: u32 = 0o2000;
+/// Set UID bit.
+pub const S_ISUID: u32 = 0o4000;
+/// File mode is a fifo / named pipe.
+pub const S_IFIFO: u32 = 0o10000;
+/// File mode is a character device.
+pub const S_IFCHR: u32 = 0o20000;
+/// File mode indicating a directory.
+pub const S_IFDIR: u32 = 0o40000;
+/// File mode indicating a block device.
+pub const S_IFBLK: u32 = 0o60000;
+/// File mode indicating a regular file.
+pub const S_IFREG: u32 = 0o100000;
+/// File mode indicating a symbolic link.
+pub const S_IFLNK: u32 = 0o120000;
+/// File mode indicating a socket.
+pub const S_IFSOCK: u32 = 0o140000;
+
 #[cfg(unix)]
 pub fn is_executable(metadata: &std::fs::Metadata) -> bool {
     let permissions = metadata.permissions();
@@ -37,6 +76,31 @@ pub fn set_executable(_file: &mut std::fs::File) -> Result<(), std::io::Error> {
     Ok(())
 }
 
+#[cfg(unix)]
+pub fn create_symlink(
+    path: impl AsRef<Path>,
+    target: impl AsRef<Path>,
+) -> Result<(), std::io::Error> {
+    std::os::unix::fs::symlink(target, path)
+}
+
+#[cfg(windows)]
+pub fn create_symlink(
+    path: impl AsRef<Path>,
+    target: impl AsRef<Path>,
+) -> Result<(), std::io::Error> {
+    let target = target.as_ref();
+
+    // The function to call depends on the type of the target.
+    let metadata = std::fs::metadata(target)?;
+
+    if metadata.is_dir() {
+        std::os::windows::fs::symlink_dir(target, path)
+    } else {
+        std::os::windows::fs::symlink_file(target, path)
+    }
+}
+
 /// Represents an abstract location for binary data.
 ///
 /// Data can be backed by the filesystem or in memory.
@@ -50,7 +114,7 @@ impl FileData {
     /// Resolve the data for this instance.
     ///
     /// If backed by a file, the file will be read.
-    pub fn resolve(&self) -> Result<Vec<u8>, std::io::Error> {
+    pub fn resolve_content(&self) -> Result<Vec<u8>, std::io::Error> {
         match self {
             Self::Path(p) => {
                 let data = std::fs::read(p)?;
@@ -65,7 +129,15 @@ impl FileData {
     ///
     /// This ensures any file-backed data is present in memory.
     pub fn to_memory(&self) -> Result<Self, std::io::Error> {
-        Ok(Self::Memory(self.resolve()?))
+        Ok(Self::Memory(self.resolve_content()?))
+    }
+
+    /// Obtain a filesystem path backing this content.
+    pub fn backing_path(&self) -> Option<&Path> {
+        match self {
+            Self::Path(p) => Some(p.as_path()),
+            Self::Memory(_) => None,
+        }
     }
 }
 
@@ -97,9 +169,13 @@ impl From<&[u8]> for FileData {
 #[derive(Clone, Debug, PartialEq)]
 pub struct FileEntry {
     /// The content of the file.
-    pub data: FileData,
+    data: FileData,
+
     /// Whether the file is executable.
-    pub executable: bool,
+    executable: bool,
+
+    /// Indicates that this file is a link pointing to the specified path.
+    link: Option<PathBuf>,
 }
 
 impl TryFrom<&Path> for FileEntry {
@@ -112,16 +188,95 @@ impl TryFrom<&Path> for FileEntry {
         Ok(Self {
             data: FileData::from(path),
             executable,
+            link: None,
         })
     }
 }
 
+impl TryFrom<PathBuf> for FileEntry {
+    type Error = std::io::Error;
+
+    fn try_from(path: PathBuf) -> Result<Self, Self::Error> {
+        Self::try_from(path.as_path())
+    }
+}
+
+impl From<&FileEntry> for FileEntry {
+    fn from(entry: &FileEntry) -> Self {
+        entry.clone()
+    }
+}
+
+impl From<Vec<u8>> for FileEntry {
+    fn from(data: Vec<u8>) -> Self {
+        Self {
+            data: data.into(),
+            executable: false,
+            link: None,
+        }
+    }
+}
+
+impl From<&[u8]> for FileEntry {
+    fn from(data: &[u8]) -> Self {
+        Self {
+            data: data.into(),
+            executable: false,
+            link: None,
+        }
+    }
+}
+
 impl FileEntry {
+    /// Construct a new instance given data and an executable bit.
+    pub fn new_from_data(data: impl Into<FileData>, executable: bool) -> Self {
+        Self {
+            data: data.into(),
+            executable,
+            link: None,
+        }
+    }
+
+    /// Construct a new instance referencing a path.
+    pub fn new_from_path(path: impl AsRef<Path>, executable: bool) -> Self {
+        Self {
+            data: path.as_ref().into(),
+            executable,
+            link: None,
+        }
+    }
+
+    /// Obtain the [FileData] backing this instance.
+    pub fn file_data(&self) -> &FileData {
+        &self.data
+    }
+
+    /// Whether the file is executable.
+    pub fn is_executable(&self) -> bool {
+        self.executable
+    }
+
+    /// Set whether the file is executable.
+    pub fn set_executable(&mut self, v: bool) {
+        self.executable = v;
+    }
+
+    /// Resolve the data constituting this file.
+    pub fn resolve_content(&self) -> Result<Vec<u8>, std::io::Error> {
+        self.data.resolve_content()
+    }
+
+    /// Obtain the target of a link, if this is a link entry.
+    pub fn link_target(&self) -> Option<&Path> {
+        self.link.as_deref()
+    }
+
     /// Obtain a new instance guaranteed to have file data stored in memory.
     pub fn to_memory(&self) -> Result<Self, std::io::Error> {
         Ok(Self {
             data: self.data.to_memory()?,
             executable: self.executable,
+            link: self.link.clone(),
         })
     }
 
@@ -133,10 +288,15 @@ impl FileEntry {
             .ok_or(FileManifestError::NoParentDirectory)?;
 
         std::fs::create_dir_all(parent)?;
-        let mut fh = std::fs::File::create(&dest_path)?;
-        fh.write_all(&self.data.resolve()?)?;
-        if self.executable {
-            set_executable(&mut fh)?;
+
+        if let Some(link) = &self.link {
+            create_symlink(dest_path, link)?;
+        } else {
+            let mut fh = std::fs::File::create(&dest_path)?;
+            fh.write_all(&self.resolve_content()?)?;
+            if self.executable {
+                set_executable(&mut fh)?;
+            }
         }
 
         Ok(())
@@ -146,8 +306,8 @@ impl FileEntry {
 /// Represents a virtual file, with an associated path.
 #[derive(Clone, Debug, PartialEq)]
 pub struct File {
-    pub path: PathBuf,
-    pub entry: FileEntry,
+    path: PathBuf,
+    entry: FileEntry,
 }
 
 impl TryFrom<&Path> for File {
@@ -171,13 +331,23 @@ impl From<File> for FileEntry {
 
 impl File {
     /// Create a new instance from a path and `FileEntry`.
-    pub fn new(path: impl AsRef<Path>, entry: FileEntry) -> Self {
+    pub fn new(path: impl AsRef<Path>, entry: impl Into<FileEntry>) -> Self {
         Self {
             path: path.as_ref().to_path_buf(),
-            entry,
+            entry: entry.into(),
         }
     }
 
+    /// The path of this instance.
+    pub fn path(&self) -> &Path {
+        &self.path
+    }
+
+    /// The [FileEntry] holding details about this file.
+    pub fn entry(&self) -> &FileEntry {
+        &self.entry
+    }
+
     /// Obtain an instance that is guaranteed to be backed by memory.
     pub fn to_memory(&self) -> Result<Self, std::io::Error> {
         Ok(Self {
@@ -192,6 +362,12 @@ impl File {
     }
 }
 
+impl AsRef<Path> for File {
+    fn as_ref(&self) -> &Path {
+        &self.path
+    }
+}
+
 #[derive(Debug)]
 pub enum FileManifestError {
     IllegalRelativePath(String),
@@ -199,6 +375,7 @@ pub enum FileManifestError {
     NoParentDirectory,
     IoError(std::io::Error),
     StripPrefix(std::path::StripPrefixError),
+    LinkNotAllowed,
 }
 
 impl std::fmt::Display for FileManifestError {
@@ -213,6 +390,7 @@ impl std::fmt::Display for FileManifestError {
             Self::NoParentDirectory => f.write_str("could not resolve parent directory"),
             Self::IoError(inner) => inner.fmt(f),
             Self::StripPrefix(inner) => inner.fmt(f),
+            Self::LinkNotAllowed => f.write_str("links are not allowed on this FileManifest"),
         }
     }
 }
@@ -257,9 +435,19 @@ pub fn normalize_path(path: &Path) -> Result<PathBuf, FileManifestError> {
 #[derive(Clone, Debug, Default, PartialEq)]
 pub struct FileManifest {
     files: BTreeMap<PathBuf, FileEntry>,
+    /// Whether to allow storage of links.
+    allow_links: bool,
 }
 
 impl FileManifest {
+    /// Create a new instance that allows the storage of links.
+    pub fn new_with_links() -> Self {
+        Self {
+            files: BTreeMap::new(),
+            allow_links: true,
+        }
+    }
+
     /// Whether the instance has any files entries.
     pub fn is_empty(&self) -> bool {
         self.files.is_empty()
@@ -317,8 +505,13 @@ impl FileManifest {
         path: impl AsRef<Path>,
         entry: impl Into<FileEntry>,
     ) -> Result<(), FileManifestError> {
-        self.files
-            .insert(normalize_path(path.as_ref())?, entry.into());
+        let entry = entry.into();
+
+        if entry.link.is_some() && !self.allow_links {
+            return Err(FileManifestError::LinkNotAllowed);
+        }
+
+        self.files.insert(normalize_path(path.as_ref())?, entry);
 
         Ok(())
     }
@@ -335,6 +528,21 @@ impl FileManifest {
         Ok(())
     }
 
+    /// Add a symlink to the manifest.
+    pub fn add_symlink(
+        &mut self,
+        manifest_path: impl AsRef<Path>,
+        link_target: impl AsRef<Path>,
+    ) -> Result<(), FileManifestError> {
+        let entry = FileEntry {
+            data: vec![].into(),
+            executable: false,
+            link: Some(link_target.as_ref().to_path_buf()),
+        };
+
+        self.add_file_entry(manifest_path, entry)
+    }
+
     /// Merge the content of another manifest into this one.
     ///
     /// All entries from the other manifest are overlayed into this manifest while
@@ -485,19 +693,32 @@ impl FileManifest {
 
         self.materialize_files(dest)
     }
+
+    /// Ensure the content of all entries is backed by memory.
+    pub fn ensure_in_memory(&mut self) -> Result<(), std::io::Error> {
+        for entry in self.files.values_mut() {
+            entry.data = entry.data.to_memory()?;
+        }
+
+        Ok(())
+    }
 }
 
 #[cfg(test)]
 mod tests {
-    use super::*;
+    use {super::*, tempfile::TempDir};
+
+    #[cfg(unix)]
+    fn temp_dir() -> std::io::Result<TempDir> {
+        tempfile::Builder::new()
+            .prefix("tugger-file-manifest-test-")
+            .tempdir()
+    }
 
     #[test]
     fn test_add_file_entry() -> Result<(), FileManifestError> {
         let mut m = FileManifest::default();
-        let f = FileEntry {
-            data: FileData::from(vec![42]),
-            executable: false,
-        };
+        let f = FileEntry::new_from_data(vec![42], false);
 
         m.add_file_entry(Path::new("foo"), f.clone())?;
 
@@ -515,20 +736,8 @@ mod tests {
         let mut m = FileManifest::default();
 
         let files = vec![
-            File {
-                path: Path::new("foo").into(),
-                entry: FileEntry {
-                    data: FileData::from(vec![42]),
-                    executable: false,
-                },
-            },
-            File {
-                path: Path::new("dir0/file0").into(),
-                entry: FileEntry {
-                    data: FileData::from(vec![42]),
-                    executable: false,
-                },
-            },
+            File::new("foo", vec![42]),
+            File::new("dir0/file0", vec![42]),
         ];
 
         m.add_files(files.into_iter())?;
@@ -541,10 +750,7 @@ mod tests {
     #[test]
     fn test_add_bad_path() -> Result<(), FileManifestError> {
         let mut m = FileManifest::default();
-        let f = FileEntry {
-            data: FileData::from(vec![]),
-            executable: false,
-        };
+        let f = FileEntry::new_from_data(vec![], false);
 
         let res = m.add_file_entry(Path::new("../etc/passwd"), f.clone());
         let err = res.err().unwrap();
@@ -563,13 +769,42 @@ mod tests {
         Ok(())
     }
 
+    #[test]
+    fn add_link_rejected() -> Result<(), FileManifestError> {
+        let mut m = FileManifest::default();
+        let mut f = FileEntry::from(vec![42]);
+        f.link = Some("/etc/passwd".into());
+
+        let res = m.add_file_entry("foo", f);
+        let err = res.err().unwrap();
+        assert!(matches!(err, FileManifestError::LinkNotAllowed));
+
+        Ok(())
+    }
+
+    #[cfg(unix)]
+    #[test]
+    fn symlink_unix() -> Result<(), FileManifestError> {
+        let mut m = FileManifest::new_with_links();
+        m.add_symlink("etc", "/etc")?;
+
+        let td = temp_dir()?;
+
+        m.materialize_files(td.path())?;
+
+        let p = td.path().join("etc");
+        let metadata = std::fs::symlink_metadata(&p)?;
+
+        assert_ne!(metadata.permissions().mode() & S_IFLNK, 0);
+        assert_eq!(std::fs::read_link(&p)?, PathBuf::from("/etc"));
+
+        Ok(())
+    }
+
     #[test]
     fn test_relative_directories() -> Result<(), FileManifestError> {
         let mut m = FileManifest::default();
-        let f = FileEntry {
-            data: FileData::from(vec![]),
-            executable: false,
-        };
+        let f = FileEntry::new_from_data(vec![], false);
 
         m.add_file_entry(Path::new("foo"), f.clone())?;
         let dirs = m.relative_directories();
@@ -588,10 +823,7 @@ mod tests {
     #[test]
     fn test_resolve_directories() -> Result<(), FileManifestError> {
         let mut m = FileManifest::default();
-        let f = FileEntry {
-            data: vec![].into(),
-            executable: false,
-        };
+        let f = FileEntry::new_from_data(vec![], false);
 
         m.add_file_entry(Path::new("foo"), f.clone())?;
         m.add_file_entry(Path::new("dir1/dir2/foo"), f)?;
@@ -611,10 +843,7 @@ mod tests {
 
     #[test]
     fn test_entries_by_directory() -> Result<(), FileManifestError> {
-        let c = FileEntry {
-            data: vec![42].into(),
-            executable: false,
-        };
+        let c = FileEntry::new_from_data(vec![42], false);
 
         let mut m = FileManifest::default();
         m.add_file_entry(Path::new("root.txt"), c.clone())?;
@@ -713,10 +942,7 @@ mod tests {
 
     #[test]
     fn test_entries_by_directory_windows() -> Result<(), FileManifestError> {
-        let c = FileEntry {
-            data: vec![42].into(),
-            executable: false,
-        };
+        let c = FileEntry::new_from_data(vec![42], false);
 
         let mut m = FileManifest::default();
         m.add_file_entry(Path::new("root.txt"), c.clone())?;
